-- Stolen and adapted from "Genuinely Functional User Interfaces", by
-- Antony Courtney and Conal Elliott.

-- http://www.apocalypse.org/pub/u/antony/pubs/genuinely-functional-guis.pdf

paddle :: Double -> G.Rectangle2DDouble
paddle xpos = G.rectangle (xpos - 25) 200 50 10 
-- The paddleball game, capable of playing one game 
-- Output signal is an Event Source that occurs 
-- when the game ends. 

pball :: Double -> GUI () (Maybe ()) 
pball vel = proc (inpS,_) -> do 
  rec 
      xvel >- integral    -> xi 
      let xpos = 30 + xi 
      yvel >- integral    -> yi 
      let ypos = 30 + yi 
          ballS = ell (xpos-12.5) (ypos-12.5) 25 25 
          ballPicS = G.shapePic ballS `G.withColor` G.yellow 
      (xpos > 175) || (xpos < 30) >- implies    -> xbounce 
      (ypos < 30) || hitPaddle >- implies       -> ybounce 
      let hitPaddle = intersects ballS paddleS 
      ebind negate xbounce >- stepAccum vel     -> xvel 
      ebind negate ybounce >- stepAccum vel     -> yvel 
      inpS >- mouseB -> mpos 
      let paddleS = paddle (G.pointX mpos) 
          paddlePicS = G.shapePic paddleS `G.withColor` G.green 
      ypos > 200 >- implies -> gameOver 
      let gamePic = G.box (walls `G.over` paddlePicS `G.over` 
                           ballPicS) gameBox 
  (gamePic,gameOver) >- returnA

-- A version that uses vectors instead of scalars:

pball2 :: Double -> GUI () (Maybe ()) 
pball2 speed = proc () -> do 
  rec 
      vel >- integral                                -> p
      let pos = 30 + p 
          ballS = ell (pos - 12.5) (point 25 25)
          ballPicS = G.shapePic ballS 
                     `G.withColor` G.yellow 
      (xc pos > 175) || (xc pos < 30) >- implies     -> xbounce 
      (yc pos < 30)  || hitPaddle     >- implies     -> ybounce 
      yc pos > 200                    >- implies     -> gameOver 
      (-xbounce)  (-ybounce)         >- point       -> bounce
      let hitPaddle = intersects ballS paddleS     
      bounce >- stepAccum speed                      -> vel
      () >- mouseB -> mpos 
      let paddleS = paddle (G.pointX mpos) 
          paddlePicS = G.shapePic paddleS 
                       `G.withColor` G.green 
      let gamePic = G.box (walls `G.over` 
                           paddlePicS `G.over` 
                           ballPicS) gameBox 
      () >- render gamePic
  gameOver >- returnA
