-- -*- haskell-hugs-program-args: ("+." "-98") -*-
module ProdArrowInstances where

import ProdArrows

instance Arrow (->) where
  arr f = f
  f >>> g = g . f


newtype StateArrow s a b = ST ((a,s) -> (b,s))
instance Arrow (StateArrow s) where
  arr f = ST (first f)
  ST f >>> ST g = ST (f >>> g)

instance ArrowUnit (->) where
  aunit = \ _ -> ()

instance ProdArrow (->) (,) where
  f &&& g = \x -> (f x, g x)
  afst = fst
  asnd = snd
  combine = id

instance ArrowApply (->) (,) where
  app (f,a) = f a

instance ArrowLoop (->) (,) where
  loop f b = c where
     ~(c,d) = f (b,d)

instance ArrowChoice (->) where
  (|||) = either

instance (Show a,Show b) => Show (a->b) where show f = "<function>"
data SP a b = Put b (SP a b) | Get (a -> SP a b) deriving Show

nullSP = Get $ \a -> nullSP

(.>) = (>)
(<.) = flip (.>)

a  > Get f = f a
a  > Put b f = Put b $ a > f

instance Arrow SP where
  arr f = Get $ \a -> Put (f a) $ arr f
  f       >>> Put c g = Put c (f >>> g)
  Put b f >>> Get g   = f >>> (g b)
  Get f   >>> Get g   = Get $ \a -> f a >>> Get g

instance ArrowUnit SP where
  aunit = Put () nullSP

instance ProdArrow SP Either where
  Put c f &&& g = Put (Left c) $ f &&& g
  f &&& Put c' g = Put (Right c') $ f &&& g
  Get f &&& Get g = Get $ \x -> f x &&& g x

  afst = Get $ (\x -> Put x afst) `either` (\x -> afst)
  asnd = Get $ (\x -> asnd) `either` (\x -> Put x asnd)

  combine = comb Nothing Nothing where
    comb a b = Get $ (\x -> out (Just x) b) `either` (\x -> out a (Just x))
    out (Just a) (Just b) = Put (a,b) $ comb (Just a) (Just b)
    out a b = comb a b

combine' = comb Nothing where
    comb Nothing = Get $ (comb . Just . Left) `either` (comb . Just . Right)
    comb (Just (Left x))  = Get $ (comb . Just . Left) `either` 
                                  (\y -> Put (x,y) $ combine')
    comb (Just (Right y)) = Get $ (\x -> Put (x,y) $ combine') `either` 
                                  (comb . Just . Right)


instance ArrowApply SP Either where
  app  = Get $ appI Nothing where
      appI _ (Left f) = appO f
      appI (Just f) (Right a) = appO (f a)
      appI Nothing (Right a) = app
      appO (Put a f) = Put a $ appO f
      appO (Get f) = Get $ appI (Just f)

instance ArrowLoop SP Either where
  loop (Put (Left c) f) = Put c $ loop f
  loop (Put (Right d) f) = loop (Right d > f)
  loop (Get f) = Get $ (loop . f . Left)

instance ArrowChoice SP where
  f ||| g = (f *** g) >>> arr stripEither

stripEither (Left  x) = x
stripEither (Right x) = x