ToDo: Are initial values for SrcLoc/current column correct?

> module Main (testLexer, main) where

> import HsLexer
> import HsParseMonad
> import HsParser
> import HsSyn
> import TranslateArrows(translateArrows)
> import HsPretty
> import System
> import GetOpt

> data Flag = LexOnlyLength          -- print number of tokens only
>           | LexOnlyRev             -- print tokens in reverse order
>           | LexOnly                -- print tokens
>           | ParseLength            -- print number of declarations only
>           | ParseInternal          -- print abstract syntax in internal format
>           | ParsePretty PPLayout   -- pretty print in this style
>           | Help                   -- give short usage info

> usage = "usage: hsparser [option] [filename]\n"

> options =
>    [ Option ['n']  ["numtokens"] (NoArg LexOnlyLength) "print number of tokens only",
>      Option ['r']  ["revtokens"] (NoArg LexOnlyRev)    "print tokens in reverse order",
>      Option ['t']  ["tokens"]    (NoArg LexOnly)       "print tokens",
>      Option ['d']  ["numdecls"]  (NoArg ParseLength)   "print number of declarations only",
>      Option ['a']  ["abstract"]  (NoArg ParseInternal) "print abstract syntax in internal format",
>      Option ['p']  ["pretty"]    (OptArg style "STYLE")   "pretty print in STYLE[(o)ffside|(s)emicolon|(i)nline|(n)one](default = offside)",
>      Option ['h','?'] ["help"]   (NoArg Help)          "display this help and exit"]

> style :: Maybe String -> Flag
> style Nothing = ParsePretty PPOffsideRule
> style (Just s) = ParsePretty $ case s of
>				    "o" -> PPOffsideRule
>				    "offside" -> PPOffsideRule
>				    "s" -> PPSemiColon
>				    "semicolon" -> PPSemiColon
>				    "i" -> PPInLine
>				    "inline" -> PPInLine
>				    "n" -> PPNoLayout
>				    "none" -> PPNoLayout
>				    _ -> PPOffsideRule

> main :: IO ()
> main = do cmdline <- getArgs
>           mainHugs cmdline

> mainHugs :: [String] -> IO ()
> mainHugs cmdline =
>    case getOpt Permute options cmdline of
>       (flags, args, [])     -> do inp <- case args of
>                                             []  -> getContents
>                                             [f] -> readFile f
>                                             _   -> error usage
>                                   putStrLn (handleFlag (getFlag flags) inp)
>       (_,     _,    errors) -> error (concat errors ++ usageInfo usage options)

> getFlag :: [Flag] -> Flag
> getFlag []  = ParsePretty PPOffsideRule
> getFlag [f] = f
> getFlag _   = error usage

> handleFlag :: Flag -> String -> String
> handleFlag LexOnlyLength = show . numToks . testLexerRev
> handleFlag LexOnlyRev    = show . testLexerRev
> handleFlag LexOnly       = show . testLexer
> handleFlag ParseLength   = show . allLengths . testParser
>    where allLengths (HsModule _ _ imp d) = length imp + length d
> handleFlag ParseInternal = show . testParser
> handleFlag (ParsePretty layout) = renderWithMode defaultMode{layout=layout} 
>				    . ppHsModule . translateArrows . testParser
> handleFlag Help          = const $ usageInfo ("A simple test program for *The Haskell Parser*\n" ++ usage) options

> numToks :: ParseResult [Token] -> Int
> numToks (Ok _ toks)  = length toks
> numToks (Failed err) = error ("Huh? " ++ err)

> testLexerRev :: String -> ParseResult [Token]
> testLexerRev s = loop [] s (SrcLoc 1 0) 1 []   -- this magic should be abstracted out...
>   where loop toks =
>	     lexer (\t -> case t of 
>				EOF -> returnP toks
>				_   -> loop (t:toks))

> testLexer :: String -> ParseResult [Token]
> testLexer s = loop [] s (SrcLoc 1 0) 1 []
>   where loop toks =
>	     lexer (\t -> case t of 
>				EOF -> returnP (reverse toks) -- space leak?
>				_   -> loop (t:toks))

> testParser :: String -> HsModule
> testParser s = case parse s (SrcLoc 1 1) 0 [] of
>		    Ok state e -> e
>		    Failed err -> error err
