-- -*- haskell-hugs-program-args: ("+." "-98") -*-
module TranslateArrows where

import List(intersect)
import HsSyn
import SynOps

deb' s a b = deb (s ++ show a) b

translateArrows x = mapExp (const (compileArrows apunit)) [] (desugarDo x)

desugarDo :: MapExp a => a -> a
desugarDo = mapExpPart (const desugarExpPart) []
  where
 desugarDoDecls = mapExpDecls (const desugarDo) []

 desugarExpPart e@(HsArrowProc p c) = Just (HsArrowProc p (desugarCmd c))
 desugarExpPart e                   = Nothing

 desugarCmd (a :-< b) = desugarDo a :-< desugarDo b
 desugarCmd (ArForm a cs) = ArForm (desugarDo a) (map desugarCmd cs)
 desugarCmd (ArInfixOp c1 e c2) = ArInfixOp (desugarCmd c1) 
                                            (desugarDo e) 
                                            (desugarCmd c2)
 desugarCmd (ArKappa p c) = ArKappa p (desugarCmd c)
 desugarCmd (ArExp e) = ArExp (desugarDo e)
 desugarCmd (ArLet ds c) = ArLet (desugarDoDecls ds) (desugarCmd c)
 desugarCmd (ArDo (ss,c)) = desugarStmt ss c

 desugarStmt :: [ArrowStmt] -> ArrowCmd -> ArrowCmd
 desugarStmt [] c = desugarCmd c
 desugarStmt (ArBind p c1:ss) c = ArInfixOp (desugarCmd c1) (HsVar bindVar)
                                   (ArKappa p (desugarStmt ss c))
 desugarStmt (ArLetStmt ds:ss) c = ArLet (desugarDoDecls ds) (desugarStmt ss c)
 desugarStmt (ArRec a:b) c = desugarStmt (r:b) c
     where r = ArBind (coprod pB)
                      (ArForm (HsVar loopVar) [ArKappa (coprod pA) $
                        desugarStmt a (a2h returnA :-< 
                                       (p2e (coprod pB :*: coprod pA)))])
           definedA = defined a
           pA = definedA `intersect` (fvArrowStmts (a,dummy))
                 where dummy = ArExp (HsTuple [])
           pB = definedA `intersect` (fvArrowStmts (b,c))

punit = HsPVar (UnQual "()")
apunit = ArHsPat punit

coprod [] = apunit
coprod l = foldr1 (:*:) (map (ArHsPat . HsPVar) l)

tuple x y = HsPTuple [x,y]

prod [] = punit
prod l = foldr1 tuple $ map HsPVar l

compileArrows p = mapExpPart (\n -> compileArrow (zapVar n p)) []

compileArrow _ (HsArrowProc p c) = Just (paren (a2h (compileProc p c)))
compileArrow p e = case h2a e [] of AExp _ -> Nothing
                                    a      -> Just (a2h a)

compileProd :: ArrowPat -> HsExp -> Arrow
compileProd p (e1 :* e2)  = compileProd p e1 &&& compileProd p e2
compileProd p (HsParen e) = compileProd p e
compileProd p e           = compileProc p (a2h returnA :-< e)

isProd (_ :* _)    = True
isProd (HsParen e) = isProd e
isProd _           = False

{-
proc (x :*: y) -> arr id -< x = afst >>> proc x -> arr id -< x
proc (x :*: y) -> \z -> arr id -< y+z = proc (z :*: (x *: y)) -> arr id -< y +z
-}

compileProc :: ArrowPat -> ArrowCmd -> Arrow
compileProc p e@(ArKappa _ _) = cP p e
compileProc (p1 :*: p2) e 
   | novp1 && not novp2 = asnd >>> compileProc p2 e
   | novp2 && not novp1 = afst >>> compileProc p1 e
     where novp1 = null (fve `intersect` fv p1)
           novp2 = null (fve `intersect` fv p2)
           fve = fv e
compileProc p e = cP p e
cP p (e1 :-< e2)
   | nullp p1  = a2 >>> h2a e1' []
   | otherwise = (a1 &&& a2) >>> app
      where p1             = prod $ lfv p e1
            p2             = prod $ lfv p e2
            e1'            = compileArrows p e1
            e2'            = compileArrows p e2
            a2 | isProd e2 = compileProd p e2
               | otherwise = proj p (ArHsPat p2) >>> (p2 ==> e2')
            a1             = proj p (ArHsPat p1) >>> (p1 ==> e1')

cP p (ArForm e [a,b]) | e == HsVar bindVar = compileProc p a `bind`
                                             compileProc p b
cP p (ArForm e ss) = h2a (compileArrows p e) 
                         (map (a2h . compileProc p) ss)
cP p (ArInfixOp c1 e c2) = compileProc p (ArForm e [c1,c2])
cP p (ArKappa p' c) = compileProc (zapVar (fv p') p :*: p') c
cP p (ArExp e) = h2a (compileArrows p e) []
cP p e@(ArLet ds c) = AExp (HsLet ds (a2h (compileProc 
                                           (zapVar (defined ds) p) c)))
cP p e@(ArDo _) = error ("compileProc: "++ show e)

a2h :: Arrow -> HsExp
a2h (a :&&& b) = HsParen (HsInfixApp (a2h a) hand (a2h b))
a2h (Acomp l)  = HsParen (foldr1 (\a r -> HsInfixApp a hcomp r) (map a2h l))
a2h (p :-> e)  = HsApp harr (lambda p e)
a2h (Arr e)    = HsApp harr (paren e)
a2h (Afst)     = hafst
a2h (Asnd)     = hasnd
a2h (Combine)  = hcombine
a2h (Aunit)    = hunit
a2h (App)      = happ
a2h (AExp h)   = paren h

h2a :: HsExp -> [HsExp] -> Arrow
h2a (HsApp f e) l                           = h2a f (e:l)
h2a (HsParen a) l                           = h2a a l
h2a a [HsParen b]                           = h2a a [b]
h2a a [HsLambda [p] b]    | a == harr       = p ==> b
h2a a [b]                 | a == harr       = Arr b
h2a (HsInfixApp b a c) [] | a == hand       = h2a b [] &&& h2a c []
h2a (HsInfixApp b a c) [] | a == hcomp      = h2a b [] >>> h2a c []
h2a (HsInfixApp b a c) [] | a == hstar      = h2a b [] *** h2a c []
h2a a []                  | a == returnAVar = Arr idVar
h2a a []                  | a == hafst      = afst
h2a a []                  | a == hasnd      = asnd
h2a a []                  | a == hcombine   = combine
h2a a []                  | a == hunit      = aunit
h2a a []                  | a == happ       = app
h2a a as                                    = AExp (foldl HsApp (paren a) 
                                                                (map paren as))

paren e@(HsInfixApp _ _ _) = HsParen e
paren e@(HsApp _ _) = HsParen e
paren e@(HsNegApp _) = HsParen e
paren e@(HsLambda _ _) = HsParen e
paren e@(HsLet _ _) = HsParen e
paren e@(HsIf _ _ _) = HsParen e
paren e@(HsCase _ _) = HsParen e
paren e@(HsDo _) = HsParen e
paren e@(HsExpTypeSig _ _ _) = HsParen e
paren e@(HsArrowProc _ _) = HsParen e
paren e@(_ :* _) = HsParen e
paren e = e

arrop = HsVar . arrowMod

hand     = arrop "&&&"
hcomp    = arrop ">>>"
hstar    = arrop "***"
hafst    = arrop "afst"
hasnd    = arrop "asnd"
hcombine = arrop "combine"
hunit    = arrop "aunit"
happ     = arrop "app"
harr     = arrop "arr"
returnAVar = arrop "returnA"

data Arrow = Arrow :&&& Arrow
           | Acomp [Arrow] -- at least two elements
           | HsPat :-> HsExp -- arr (\p -> e)
           | Arr HsExp
           | Afst
           | Asnd
           | Combine
           | Aunit
           | App
           | AExp HsExp
  deriving (Show,Eq)

nullp p | p == punit = True
nullp _ = False


isId (Arr i)   | i == idVar = True
isId (p :-> e) | p2e p == e = True
isId _ = False

isPure :: Arrow -> Bool
isPure (a :&&& b) = isPure a && isPure b
isPure (Acomp l)  = all isPure l
isPure (_ :-> _)  = True
isPure (Arr _)    = True
isPure (AExp _)   = False
isPure App        = False
isPure _          = True

afst = Afst
asnd = Asnd
combine | prodIsPair = returnA
        | otherwise  = Combine
aunit = Aunit
app = App

prodIsPair           = False
composeIsCommutative = True

a1 *** a2 = (afst >>> a1) &&& (asnd >>> a2)

a1 >>> a2 = aComp (compose (deAcomp a1 ++ deAcomp a2))

deAcomp (Acomp l)     = l
deAcomp a | isId a    = []
          | otherwise = [a]

aComp [a] = a
aComp []  = returnA
aComp l   = Acomp l

compose ((p1 :-> p2):(p3 :-> p4):as) 
  | p2 == p2e p3 = compose ((p1 ==> p4):as)
  | otherwise    = compose ((p1 ==> HsApp (lambda p3 p4) (paren p2)):as)
compose ((p1 :-> p2):Arr e:as) 
  | null (fv e `intersect` fv p1) = compose ((p1 ==> HsApp e (paren p2)):as)
  | otherwise                     = compose (arrComp (lambda p1 p2) e:as)
compose (Arr e : (p1 :-> p2) :as)    = compose (arrComp e (lambda p1 p2):as)
compose (Arr e:Arr f:as)          = compose (arrComp e f:as)
compose ((a1 :&&& a2) : Afst :as) | isPure a2 = compose (deAcomp a1++as)
compose ((a1 :&&& a2) : Asnd :as) | isPure a1 = compose (deAcomp a2++as) -- *
compose ((a1 :&&& a2) : (b1 :&&& b2) :as) -- * (at least with the next)
   | hdIsAsnd b1 && hdIsAfst b2
   = compose ((a2 :&&& a1) : (compTail b1 *** compTail b2) :as)
compose ((a1 :&&& a2) : (b1 :&&& b2) :as)
   | hdIsAfst b1 && hdIsAsnd b2 &&
     (isPure b1 && isPure b2   -- *
      || composeIsCommutative) -- *
   = compose ((a1 >>> compTail b1) &&& (a2 >>> compTail b2):as)
compose (a1:a2:as) = a1 : compose (a2:as)
compose l = l

hdIsAfst Afst = True
hdIsAfst (Acomp (Afst:_)) = True
hdIsAfst _ = False

hdIsAsnd Asnd = True
hdIsAsnd (Acomp (Asnd:_)) = True
hdIsAsnd _ = False

compTail (Acomp (_:l)) = aComp l
compTail _ = returnA

arrComp e f = Arr (HsInfixApp (paren e) funComp (paren f))
lambda p1 p2 = HsParen (HsLambda [p1] p2)

-- *: case found when optimization makes a difference

Afst &&& Asnd = returnA -- *
Acomp (a1:b) &&& Acomp (a2:c) | a1 == a2 && isPure a1 -- *
                                  = a1 >>> (aComp b &&& aComp c)
a1 &&& a2 | a1 == a2 && isPure a1 = a1 >>> (returnA :&&& returnA)
a1 &&& a2 = a1 :&&& a2

s1 ==> s2 | p2e s1 == s2 = returnA
          | otherwise    = s1 :-> s2

returnA = Arr idVar

u `bind` f = (returnA &&& u) >>> f

proj :: ArrowPat -> ArrowPat -> Arrow
proj g1 g2 | g1 == g2 = returnA
proj g1 (g2 :*: g3) = proj g1 g2 &&& proj g1 g3
proj g1 (ArPParen g2) = proj g1 g2
proj g (ArHsPat p) = projs g p

projs (g1 :*: g2) p | nullp p1 && 
                      nullp p2  = aunit
                    | nullp p2  = afst >>> projs g1 p
                    | nullp p1  = asnd >>> projs g2 p
                    | otherwise = (projs g1 p1 *** projs g2 p2) >>> 
                                  combine >>> (tuple p1 p2 ==> p2e p)
          where p1 = prod (fv g1 `intersect` fvp)
                p2 = prod (fv g2 `intersect` fvp)
                fvp = fv p
projs (ArPParen g1) p = projs g1 p
projs (ArHsPat p1) p2 = projss p1 p2

projss p p' | p' == punit = aunit
projss p p' = p ==> p2e p'

lfv p e = fv e `intersect` fv p

--modQ s = Qual (Module s)
modQ s = UnQual

preludeMod = modQ "Prelude"

funComp = HsVar (modQ "Prelude" ".")
idVar = HsVar (modQ "Prelude" "id")

arrowMod = modQ "Arrows"

loopVar = arrowMod "loop"
bindVar = arrowMod "bind"

--

{-
tcmd ::= {tcmd,...,tcmd} | exp

T (proc p -> {tcmd1,...,tcmdn}) = T (proc p -> tcmd1) &&& ... &&& 
                                  T (proc p -> tcmdn)
T (proc p -> exp) = proc p -> returnA exp

proc {(x,y),(u,v)} -> cmd (x,y)

     combine :: a {b,c} (b,c)
       (identity for products)

     proc p -> e1 >- e2   ==> proj_p_p1 >>> arr (\pair[p1] -> e1) >>> e2
           where p1 = lfv(p,e1) 
                      lfv(p,e2) empty

     proc p -> e1 >- e2   ==> (proj_p_p2 >>> arr (\pair[p2] -> e2)) &&&
                              (proj_p_p1 >>> arr (\pair[p1] -> e1)) >>> app
           where p1 = lfv(p,e1) (() if RHS empty), 
                 p2 = lfv(p,e2)

The p2==() case is dubious.

proj_tp_{tp1,tp2} = proj_tp_tp1 &&& proj_tp_tp2
proj_tp_() = unit :: a b ()
proj_{tp1,tp2}_p = afst >>> proj_tp1_p  | fv(tp2) \/ fv(p) null
proj_{tp1,tp2}_p = asnd >>> proj_tp2_p  | fv(tp1) \/ fv(p) null
proj_{tp1,tp2}_p = proj_tp1_p1 *** proj_tp2_p2 >>> combine >>> arr \ (p1,p2)->p
                   where pi = fv(tpi) \/ fv(p)
proj_p_p' = arr (\p -> p')

Examples:
  proj_(x,y)_(x,y) = arr \ (x,y)->(x,y) = returnA
  proj_{x,y}_(x,y) = proj_x_x *** proj_y_y >>> combine >>> arr \ (x,y)->(x,y)
                   = afst >>> returnA &&& asnd >>> returnA >>> combine
                   = combine
  proj_(x,y)_{x,y} = arr \ (x,y)->x &&& arr \ (x,y)->y
  proj_{x,y}_{x,y} = proj_{x,y}_x &&& proj_{x,y}_y =
                   = afst &&& asnd = returnA
  proj_{x,y}_{y,x} = asnd &&& afst
  proj_{x,y,z}

-}
